use std::path::Path;

use based::asset::AssetRoutes;
use based::request::{RequestContext, StringResponse};
use based::ui::components::prelude::{
    ColoredMaterialIcon, MaterialIcon, Modal, ModalOpenButton, Shell, Timeline, TimelineElement,
    Tooltip,
};
use based::ui::components::{Breadcrumb, Card, ProgressBar, Tabs};
use based::ui::primitives::div::Center;
use based::ui::primitives::flex::{Column, Row};
use based::ui::primitives::list::{CheckIcon, CheckIconRounded, CheckIconRoundedGray};
use based::ui::primitives::Optional;
use based::ui::{prelude::*, AttrExtendable};
use based::{get_pg, page};
use config::{get_prj, Project, ProjectConfig};
use invest::Investment;
use maud::{PreEscaped, Render};
use rocket::fs::NamedFile;
use rocket::request::FromSegments;
use rocket::response::Redirect;
use rocket::{get, launch, post, routes, State};
use site::gen_shell;

mod config;
mod invest;
mod site;

pub struct PathSegment {
    pub segments: Vec<String>,
}

impl PathSegment {
    pub fn to_str(&self) -> String {
        self.segments.join("/")
    }
}

impl<'r> FromSegments<'r> for PathSegment {
    type Error = ();

    fn from_segments(
        segments: rocket::http::uri::Segments<'r, rocket::http::uri::fmt::Path>,
    ) -> Result<Self, Self::Error> {
        let paths: Vec<_> = segments
            .filter_map(|x| {
                if x == "." {
                    return None;
                }

                if x == ".." {
                    return None;
                }

                Some(x.to_string())
            })
            .collect();

        Ok(PathSegment { segments: paths })
    }
}

#[get("/prj/<path..>?<tab>")]
pub async fn project_page(
    path: PathSegment,
    ctx: RequestContext,
    shell: &State<Shell>,
    c: &State<ProjectConfig>,
    tab: Option<&str>,
) -> Option<StringResponse> {
    let mut prj = c.get(path.segments.first()?);
    for p in &path.segments[1..] {
        prj = prj.as_ref()?.sub.as_ref()?.get(p);
    }

    Some(render_project(shell, ctx, prj?, c, path.to_str(), tab.unwrap_or_default()).await)
}

#[get("/static/<file>")]
async fn icon_res(file: &str) -> Option<NamedFile> {
    let path = Path::new("static").join(file);
    NamedFile::open(path).await.ok()
}

#[get("/")]
pub async fn main_page(
    c: &State<ProjectConfig>,
    ctx: RequestContext,
    shell: &State<Shell>,
) -> StringResponse {
    if c.len() == 1 {
        let keys = c.keys().collect::<Vec<_>>();
        let first = keys.first().unwrap();
        let prj = c.get(*first).unwrap();

        render_project(shell, ctx, prj, c, format!("{first}/"), "").await
    } else {
        // TODO : root project overview
        site::gen_site(c, ctx).await
    }
}

#[allow(non_snake_case)]
pub fn DetailLink(text: &str, reference: &str) -> PreEscaped<String> {
    Flex(
        Div()
            .vanish()
            .push(text.render())
            .push(Link(reference, Text(reference)))
            .render(),
    )
    .gap_x(ScreenValue::_2)
    .direction(Direction::Row)
    .render()
}

pub fn build_invest_timeline(investments: &[Investment]) -> PreEscaped<String> {
    let mut t = Timeline();

    for i in investments {
        t = t.add_element(TimelineElement::new(
            Text(&i.when),
            Row(vec![
                Text(&i.title).render(),
                Text(&format!("{:.2} $", i.amount)).render(),
                if i.payed {
                    Tooltip(CheckIconRounded(), Text("Payed")).render()
                } else {
                    CheckIconRoundedGray()
                },
            ])
            .gap(ScreenValue::_2)
            .items_center(),
            Column(vec![
                Text(&i.description).render(),
                if !i.payed {
                    Form::new(&format!("/invest/pay?id={}", i.id))
                        .add_input(FormSubmitButton(Text("Pay")))
                        .render()
                } else {
                    Nothing()
                },
            ])
            .gap(ScreenValue::_2),
        ))
    }

    t.render()
}

pub async fn invest_tab(prj: &Project, root: &str) -> PreEscaped<String> {
    let (add_invest_id, add_invest_modal) = Modal(
        "Add Investment",
        Form::new(&format!("/invest/{root}"))
            .id("new_invest")
            .add_input(TextInput("title").required().placeholder("Title"))
            .add_input(
                TextInput("amount")
                    .icon(MaterialIcon("attach_money"))
                    .required()
                    .placeholder("Amount")
                    .pattern(r#"^\d+(\.\d+)?$"#),
            )
            .add_input(
                TextArea()
                    .name("description")
                    .required()
                    .placeholder("Description"),
            ),
        |_| {
            Center(Context(
                FormSubmitButton("Create").add_attr("form", "new_invest"),
            ))
        },
    );

    let investments = Investment::get_all_investments_of(root).await;

    let t = build_invest_timeline(&investments);

    let inv = Investment::get_total_of(root).await;

    let percentage = inv.0 / inv.1;
    let percentage = (percentage * 100.0) as u8;

    Column(vec![
        add_invest_modal,
        Margin(
            Row(vec![
                Margin(Width(
                    ScreenValue::full,
                    Row(vec![
                        ColoredMaterialIcon("paid", Green::_600),
                        Text(&format!(
                            "Total Investment: {:.2} $ / {:.2} $",
                            inv.0, inv.1
                        ))
                        .color(&Green::_600)
                        .render(),
                        Width(
                            ScreenValue::full,
                            Margin(ProgressBar(percentage, true, Green::_600)).x(ScreenValue::_12),
                        )
                        .render(),
                    ])
                    .gap(ScreenValue::_2)
                    .full_center(),
                ))
                .x(ScreenValue::auto)
                .render(),
                ModalOpenButton(
                    &add_invest_id,
                    Padding(
                        Rounded(Background(MaterialIcon("add")).color(Blue::_600))
                            .size(Size::Medium),
                    )
                    .x(ScreenValue::_2)
                    .top(ScreenValue::_2),
                )
                .render(),
            ])
            .justify(Justify::Between),
        )
        .left(ScreenValue::_4)
        .render(),
        t,
    ])
    .render()
}

pub fn info_tab(prj: &Project, root: &str) -> PreEscaped<String> {
    Flex(
        Div()
            .vanish()
            .push(Rounded(
                Background(Padding(Code(&prj.description)).all(ScreenValue::_4)).color(Gray::_900),
            ))
            .push(Margin(Text("Information").xl().semibold()).top(ScreenValue::_2))
            .push(Optional(prj.website.as_deref(), |website| {
                DetailLink("Website:", website)
            }))
            .push(Optional(prj.documentation.as_deref(), |docs| {
                DetailLink("Documentation:", docs)
            }))
            .push(Optional(prj.repositories.as_deref(), |repos| {
                Column(vec![
                    Text("Repositories:").xl().semibold().render(),
                    UnorderedList()
                        .push_for_each(repos, |repo: &_| Link(repo, Text(repo)))
                        .render(),
                ])
            }))
            .push(Optional(prj.since.as_deref(), |since| {
                Text(&format!("Since: {since}"))
            }))
            .push(Optional(prj.contact.as_ref(), |contact| {
                Div()
                    .vanish()
                    .push(Text("Contact").xl().semibold())
                    .push(Text(&format!(
                        "eMail: {}",
                        contact
                            .email
                            .as_ref()
                            .map(|x| x.as_str())
                            .unwrap_or_default()
                    )))
            }))
            .push(Optional(prj.sub.as_ref(), |sub| {
                let mut prj_links = Vec::new();

                for key in sub.keys() {
                    prj_links.push(key);
                }

                Div()
                    .vanish()
                    .push(Margin(Text("Sub projects").xl().semibold()).bottom(ScreenValue::_2))
                    .push(Row(prj_links
                        .into_iter()
                        .map(|link| {
                            Width(
                                ScreenValue::fit,
                                Button(Link(
                                    &format!("/prj/{root}{link}"),
                                    ProjectIcon(
                                        &prj.sub.as_ref().unwrap().get(link).unwrap(),
                                        false,
                                    ),
                                )),
                            )
                        })
                        .collect()))
            })),
    )
    .gap(ScreenValue::_2)
    .direction(Direction::Column)
    .render()
}

fn up_to(keyword: &str, v: &[&str], separator: &str) -> String {
    let index = v.iter().position(|s| *s == keyword).unwrap_or(v.len());
    v[..index + 1].join(separator)
}

#[allow(non_snake_case)]
pub fn ProjectIcon(prj: &Project, big: bool) -> PreEscaped<String> {
    Row(vec![
        Optional(prj.icon.as_ref(), |icon| Image(icon).alt("Project Icon")),
        if big {
            Text(&prj.name)._2xl().render()
        } else {
            Text(&prj.name).render()
        },
    ])
    .gap(ScreenValue::_4)
    .full_center()
    .render()
}

pub async fn render_project(
    shell: &State<Shell>,
    ctx: RequestContext,
    prj: &Project,
    c: &State<ProjectConfig>,
    mut root: String,
    tab: &str,
) -> StringResponse {
    let title = format!("{} - Umbrella ☂️", prj.name);

    if !root.is_empty() && !root.ends_with('/') {
        root.push_str("/");
    }

    let root_paths: Vec<_> = root.split("/").filter(|x| !x.is_empty()).collect();

    let bc: Vec<_> = root_paths
        .iter()
        .map(|x| {
            let prj_path = up_to(x, &root_paths, "/");

            (
                ProjectIcon(&get_prj(c, &prj_path).unwrap(), true).0,
                format!("/prj/{}", prj_path),
            )
        })
        .collect();

    let invest_tab = invest_tab(prj, &root).await;

    page!(
        shell,
        ctx,
        title,
        Div()
            .vanish()
            .push(Width(
                ScreenValue::fit,
                Margin(Breadcrumb(bc))
                    .top(ScreenValue::_6)
                    .bottom(ScreenValue::_2)
                    .x(ScreenValue::auto)
            ))
            .push(
                Padding(
                    Tabs()
                        .active(tab)
                        .add_tab("info", Text("📜 Info").medium(), info_tab(prj, &root))
                        .add_tab("invest", Text("💵 Invest").medium(), invest_tab)
                )
                .x(ScreenValue::_6)
            )
    )
}

// todo : fav icon

#[launch]
async fn rocket() -> _ {
    let conf_path: String = std::env::args()
        .skip(1)
        .next()
        .unwrap_or("./config.toml".to_string());
    let conf = config::Config::load(&conf_path);

    let shell = gen_shell();

    sqlx::migrate!("./migrations").run(get_pg!()).await.unwrap();

    rocket::build()
        .mount_assets()
        .mount(
            "/",
            routes![
                main_page,
                icon_res,
                project_page,
                invest::post_new_invest,
                pay_investment
            ],
        )
        .manage(conf)
        .manage(shell)
}

#[post("/invest/pay?<id>")]
pub async fn pay_investment(id: i64) -> Redirect {
    let prj = Investment::pay(id).await;
    Redirect::to(format!("/prj/{}?tab=invest", prj))
}
